#ifndef _COMMON_H_
#define _COMMON_H_

#include "cuda.h"
#include "cuda_runtime_api.h"

/* Used to stop pthreads when loading CPU and/or GPU */
extern int stop_process;

typedef struct matrix_t{
    float* elements;
    uint rows;
    uint columns;
    uint stride;
} matrix;

/* Number of threads in a block's dim */
#define BLOCK_SIZE 16

#define ASSERT_FAIL 0
#define ASSERT_SUCCESS 1

/* Width of a matrix and value used for checking */
#define MATRIX_ROW_SIZE BLOCK_SIZE*16*1
#define MATRIX_VALUE 2

#define assert_cuda( err_id ) __assert_cuda(err_id,__FILE__,__LINE__)
#define assert_cu( err_id ) __assert_cu(err_id,__FILE__,__LINE__)
#define assert_gdr( err_id ) __assert_gdr(err_id,__FILE__,__LINE__)


void __assert_cuda(cudaError_t err_id, const char* file, int line); /* for runtime api*/
void __assert_cu(CUresult res_id, const char* file, int line); /* for driver api */
void __assert_gdr(int gdr_id, const char* file, int line);

/* Generates a binary to_send file used to transmit data to FPGA */
void init_to_send(const void* dataPtr, size_t size, size_t nmemb);

/* Checks if array of size floats is filled with value */
bool check_array(float* array, float value, size_t size);

/* Prints some information on device */
void deviceInformation(int device);

/* Functions to fill array with size floats of defined or random value */
void cpu_fill_array(float* array, float value, size_t size);
void cpu_fill_array_random(float* array, size_t size);

/* Uses CPU to compute average and standard deviation of array of size floats */
float cpu_average(float* array, size_t size);
float cpu_dispersion(float* array, float average, size_t size);

/* Returns an identity matrix of size row_size*row_size */
matrix identity_matrix(size_t row_size);

/* Asserts if M is identity matrix */
void check_identity_matrix(matrix M);

/* Asserts if matrix is filled with value */
void check_matrix(matrix M, int value);

/* Computes R=A*B with CPU */
void mult_matrix(matrix A, matrix B, matrix R);

/* Fills M with random values with CPU */
void fill_matrix_random(matrix M);

/* Functions to load the CPU/GPU with:
 * computation: loops a matrix multiplication
 * memory operations: loops memcpy
 */
void* cpu_load_compute(void* arg);
void* gpu_load_compute(void* arg);
void* cpu_load_memory(void* arg);
void* gpu_load_memory(void* arg);

#endif
