/* This file contains the kernels i.e. the functions to be executed on the GPU */

#include <stdio.h>
#include <stdlib.h>
#include "kernels.h"
#include "common.h"

__device__
void add_two_device(CUdeviceptr number)
{
    (* (int*) number)+=2;
}

__global__
void add_three_global(CUdeviceptr number)
{
    (* (int*) number)++;
    add_two_device(number);
}

__global__
void add_one_global(CUdeviceptr number)
{
    (* (int*) number)++;
}

__global__
void fill_array(int* array, int value, size_t size)
{
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint step = blockDim.x*gridDim.x;
    for(int i=0;x+i<size;i+=step)
	array[x+i] = value;
}

__global__
void fill_float_array(float* array, float value, size_t size)
{
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint step = blockDim.x*gridDim.x;
    for(int i=0;x+i<size;i+=step)
	array[x+i] = value;
}


extern __shared__ float _temp[];
__global__ 
void get_reduction(float* array, float* output, size_t size, size_t stride)
{
    float* temp = (float*)_temp;
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    if(x<size)
	temp[threadIdx.x]=array[x];
    __syncthreads();
    for(uint s=1;s<blockDim.x;s*=2)
    {
	x = threadIdx.x*2*s;
	if(x+s < blockDim.x)
	    temp[x]+=temp[x+s];
	__syncthreads();
    }
    if(threadIdx.x==0)
    	output[blockIdx.x+stride*gridDim.x]=temp[0];
}
__global__ 
void get_average_reduction(float* array, float* average, size_t size, size_t stride, size_t divider)
{
    float* temp = (float*)_temp;
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    if(x<size)
	temp[threadIdx.x]=array[x];
    __syncthreads();
    for(uint s=1;s<blockDim.x;s*=2)
    {
	x = threadIdx.x*2*s;
	if(x+s < blockDim.x)
	    temp[x]+=temp[x+s];
	__syncthreads();
    }
    if(threadIdx.x==0)
    	*average=temp[0]/divider;
}
__global__ 
void get_dispersion_reduction(float* array, float* average, float* output, size_t size, size_t stride, size_t divider)
{
    float* temp = (float*)_temp;
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    if(x<size)
	temp[threadIdx.x]=(array[x]-*average) * (array[x]-*average);
    __syncthreads();
    for(uint s=1;s<blockDim.x;s*=2)
    {
	x = threadIdx.x*2*s;
	if(x+s < blockDim.x)
	    temp[x]+=temp[x+s];
	__syncthreads();
    }
    if(threadIdx.x==0)
    	output[blockIdx.x+stride*gridDim.x]=temp[0]/divider;
}

void get_average(float* array, float* average, size_t size,dim3 blocks_per_grid, dim3 threads_per_block)
{
    float *output,*extend;
    size_t stride = size/(blocks_per_grid.x*threads_per_block.x);
    size_t output_size = stride*blocks_per_grid.x;
    size_t extend_size = stride;
    assert_cuda( cudaMalloc((void**)&output,output_size*sizeof(float)) );
    assert_cuda( cudaMalloc((void**)&extend,extend_size*sizeof(float)) );
    stride=0;
    while(stride*blocks_per_grid.x*threads_per_block.x < size)
    {
	get_reduction
	    <<< blocks_per_grid,threads_per_block,threads_per_block.x*sizeof(float)  >>>
	    (array+stride*blocks_per_grid.x*threads_per_block.x,
	     output,
	     blocks_per_grid.x*threads_per_block.x,
	     stride);
	stride++;
    }
    assert_cuda( cudaDeviceSynchronize() );
    get_reduction
	<<< extend_size,blocks_per_grid,blocks_per_grid.x*sizeof(float)  >>>
	(output,
	 extend,
	 output_size,
	 0);
    assert_cuda( cudaDeviceSynchronize() );
    get_average_reduction
	<<< 1,extend_size,extend_size*sizeof(float)  >>>
	(extend,
	 average,
	 extend_size,//blocks_per_grid.x,
	 0,
	 size);
    assert_cuda( cudaFree(output) );
    assert_cuda( cudaFree(extend) );
}

void get_dispersion(float* array, float* average, float* dispersion, size_t size, dim3 blocks_per_grid, dim3 threads_per_block)
{
    get_average(array,average,size,blocks_per_grid,threads_per_block);
    float *output,*extend;
    size_t stride = size/(blocks_per_grid.x*threads_per_block.x);
    size_t output_size = stride*blocks_per_grid.x;
    size_t extend_size = stride;
    assert_cuda( cudaMalloc((void**)&output,output_size*sizeof(float)) );
    assert_cuda( cudaMalloc((void**)&extend,extend_size*sizeof(float)) );
    stride=0;
    while(stride*blocks_per_grid.x*threads_per_block.x < size)
    {
	get_dispersion_reduction
	    <<< blocks_per_grid,threads_per_block,threads_per_block.x*sizeof(float)  >>>
	    (array+stride*blocks_per_grid.x*threads_per_block.x,
	     average,
	     output,
	     blocks_per_grid.x*threads_per_block.x,
	     stride,
	     size);
	stride++;
    }
    assert_cuda( cudaDeviceSynchronize() );
    get_reduction
	<<< extend_size,blocks_per_grid,blocks_per_grid.x*sizeof(float)  >>>
	(output,
	 extend,
	 output_size,
	 0);
    assert_cuda( cudaDeviceSynchronize() );
    get_reduction
      	<<< 1,extend_size,blocks_per_grid.x*sizeof(float)  >>>
    	(extend,
    	 dispersion,
    	 extend_size,//blocks_per_grid.x,
    	 0);
    assert_cuda( cudaFree(output) );
    assert_cuda( cudaFree(extend) );
}
