/* Square matrix with size multiple of BLOCK_SIZE */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include <unistd.h>
#include <sys/time.h>
#include <math.h>
#include "kernels.h"
#include "gdrapi.h"
#include <pcilib.h>
#include <pcilib/kmem.h>
#include <pcilib/bar.h>
#include "ipedma.h"

#define KMEM_DEFAULT_FLAGS      (pcilib_kmem_flags_t)(PCILIB_KMEM_FLAG_HARDWARE | PCILIB_KMEM_FLAG_PERSISTENT | PCILIB_KMEM_FLAG_EXCLUSIVE)
#define KMEM_USE_RING           PCILIB_KMEM_USE(PCILIB_KMEM_USE_USER, 1)
#define KMEM_USE_DEFAULT        PCILIB_KMEM_USE(PCILIB_KMEM_USE_USER, 2)
#define GPU_NAME_LENGTH 30
#define PAGE_SIZE 1024
#define MAX_SIZE 1024*1
#define GPU_PAGE (MAX_SIZE/PAGE_SIZE)*65536
#define VALUE 5.


int main(int argc, char* argv[])
{
    /* Initialisation of the APIs */
    assert_cu( cuInit(0) );
    gdr_t g = gdr_open();
    if( g==NULL){
	printf("Could not open gdr\n");
	exit( EXIT_FAILURE );
    }
    
    /* Manage NVIDIA GPU */
    printf("\nInitialisation of the GPU\n");
    CUdevice GPU;
    CUdevprop GPUProp;
    assert_cu( cuDeviceGet(&GPU,0) );
    assert_cu( cuDeviceGetProperties(&GPUProp,GPU) );
    char gpu_name[GPU_NAME_LENGTH] = {0};
    assert_cu (cuDeviceGetName (gpu_name, GPU_NAME_LENGTH, GPU));
    printf("GPU: %s\n", gpu_name);    
    /* Check context */
    CUcontext cuCtx;
    assert_cu( cuCtxCreate(&cuCtx,CU_CTX_MAP_HOST|CU_CTX_SCHED_AUTO,GPU) );
    assert_cu( cuCtxSetCurrent(cuCtx) );
    
    /* Allocate memory on the device, pin and map */
    uint8_t flagValueToSet = 1;
    printf("\nMemory mapping with the GPU for pages\n");
    CUdeviceptr gpuPagePtr;
    assert_cu( cuMemAlloc(&gpuPagePtr,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float)) );
    assert_cu( cuPointerSetAttribute(&flagValueToSet,CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,gpuPagePtr) );
    gdr_mh_t GPUMemHandlePage;
    assert_gdr( gdr_pin_buffer(g,gpuPagePtr,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float),0,0,&GPUMemHandlePage) );
    void* gpuPageVa;
    assert_gdr( gdr_map(g,GPUMemHandlePage,&gpuPageVa,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float)) );
    gdr_info_t pageInfo;
    assert_gdr( gdr_get_info(g,GPUMemHandlePage,&pageInfo) );
    printf("Memory mapping with the GPU for descriptors\n");
    CUdeviceptr gpuDescPtr;
    assert_cu( cuMemAlloc(&gpuDescPtr,GPU_PAGE*sizeof(float)) );       
    assert_cu( cuPointerSetAttribute(&flagValueToSet,CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,gpuDescPtr) );
    gdr_mh_t GPUMemHandleDesc;
    assert_gdr( gdr_pin_buffer(g,gpuDescPtr,GPU_PAGE*sizeof(float),0,0,&GPUMemHandleDesc) );
    void* gpuDescVa;
    assert_gdr( gdr_map(g,GPUMemHandleDesc,&gpuDescVa,GPU_PAGE*sizeof(float)) );
    gdr_info_t descInfo;
    assert_gdr( gdr_get_info(g,GPUMemHandleDesc,&descInfo) );

    /* PCI */
    printf("\nSetting up the PCI\n");
    pcilib_t* pciCtx;
    char* pciVa;
    pciCtx = pcilib_open("/dev/fpga0",PCILIB_MODEL_DETECT);
    if( pciCtx == NULL ){
	printf("Cannot open a context for pci\n");
	exit( EXIT_FAILURE );
    }
    pciVa = pcilib_resolve_bar_address(pciCtx,0, 0);
    if( pciVa == NULL ){
	printf("Cannot resolve PCI physical adress to virtual\n");
	exit( EXIT_FAILURE );
    }
    CUdeviceptr dBAR;
    assert_cu( cuMemHostRegister((void*)pciVa,128,CU_MEMHOSTREGISTER_IOMEMORY) );
    assert_cu( cuMemHostGetDevicePointer(&dBAR,(void*)pciVa, 0) );
    
    /* Config PCI for Pages*/
    pcilib_kmem_handle_t* pciHandlePage;
    pciHandlePage = pcilib_alloc_kernel_memory(pciCtx, PCILIB_KMEM_TYPE_DMA_C2S_PAGE, 1, ((PAGE_SIZE%4096)?(4096 * (1 + PAGE_SIZE/4096)):PAGE_SIZE), 4096, KMEM_USE_DEFAULT, KMEM_DEFAULT_FLAGS);
    if( pciHandlePage == NULL ){
	printf("Cannot allocate  PCI kernel memory\n");
	exit( EXIT_FAILURE );
    }
    volatile void* pciMemPtrPage;
    uintptr_t pciBusPage;
    pciMemPtrPage = (uint64_t*) pcilib_kmem_get_block_ua(pciCtx,pciHandlePage,0);
    if( pciMemPtrPage == NULL ){
	printf("Cannot get PCI pointer to kernel memory\n");
	exit( EXIT_FAILURE );
    }
    pciBusPage = pcilib_kmem_get_block_ba(pciCtx,pciHandlePage,0);
    if( pciBusPage == 0 ){
	printf("Cannot get PCI Bus address on kernel memory\n");
	exit( EXIT_FAILURE );
    }
    /* Config PCI for Desc */
    pcilib_kmem_handle_t* pciHandleDesc;
    pciHandleDesc = pcilib_alloc_kernel_memory(pciCtx,PCILIB_KMEM_TYPE_CONSISTENT, 1, 128, 4096,KMEM_USE_RING, KMEM_DEFAULT_FLAGS);
    if( pciHandleDesc == NULL ){
	printf("Cannot allocate  PCI kernel memory\n");
	exit( EXIT_FAILURE );
    }
    volatile void* pciMemPtrDesc;
    uintptr_t pciBusDesc;
    pciMemPtrDesc = (uint64_t*) pcilib_kmem_get_block_ua(pciCtx,pciHandleDesc,0);
    if( pciMemPtrDesc == NULL ){
	printf("Cannot get PCI pointer to kernel memory\n");
	exit( EXIT_FAILURE );
    }
    pciBusDesc = pcilib_kmem_get_block_ba(pciCtx,pciHandleDesc,0);
    if( pciBusDesc == 0 ){
	printf("Cannot get PCI Bus address on kernel memory\n");
	exit( EXIT_FAILURE );
    }

 
    cudaError_t err;
    matrix I,matx,result;
    dim3 blocks_per_grid(MATRIX_ROW_SIZE/BLOCK_SIZE,MATRIX_ROW_SIZE/BLOCK_SIZE);
    dim3 threads_per_block(BLOCK_SIZE,BLOCK_SIZE);
    matrix dev_I,dev_matx,dev_result;

    cudaEvent_t start,stop;
    float ms_shared;
    assert_cuda( cudaEventCreate(&start) );
    assert_cuda( cudaEventCreate(&stop) );

    I.rows = MATRIX_ROW_SIZE;
    I.columns = I.rows;
    I.stride = I.columns;
    matx.rows = MATRIX_ROW_SIZE;
    matx.columns = matx.rows;
    matx.stride = matx.columns;
    result.rows = MATRIX_ROW_SIZE;
    result.columns = result.rows;
    result.stride = result.columns;
    I.elements = (float*)malloc(I.rows*I.columns*sizeof(float));
    matx.elements = (float*)malloc(matx.rows*matx.columns*sizeof(float));
    result.elements = (float*)malloc(result.rows*result.columns*sizeof(float));

    dev_I.rows = I.rows;
    dev_I.columns = dev_I.rows;
    dev_I.stride = dev_I.columns;
    dev_matx.rows = matx.rows;
    dev_matx.columns = dev_matx.rows;
    dev_matx.rows = dev_matx.columns;
    dev_result.rows = result.rows;;
    dev_result.columns = dev_result.rows;
    dev_result.stride = dev_result.columns;
    assert_cuda( cudaMalloc((void**)&dev_I.elements,I.rows*I.columns*sizeof(float)) );
    matx.elements = (float*)gpuPageVa;
    dev_matx.elements = (float*)gpuPagePtr;
    assert_cuda( cudaMalloc((void**)&dev_result.elements,result.rows*result.columns*sizeof(float)) );

    kern_identity_matrix<<< blocks_per_grid,threads_per_block >>>(dev_I);
    err = cudaGetLastError();
    printf("%s: %s\n",cudaGetErrorName(err),cudaGetErrorString(err));
    assert_cuda( cudaMemcpy(I.elements,dev_I.elements,I.rows*I.columns*sizeof(float),cudaMemcpyDeviceToHost) );

    /* Control check */
    check_identity_matrix(I);

    
    /* FPGA */
    for(int i=0;i<1;i++){//MATRIX_ROW_SIZE;i++){
	printf("iteration %d/%d\n",i+1,MATRIX_ROW_SIZE);
	init_to_send(I.elements+i*MATRIX_ROW_SIZE,sizeof(float),MATRIX_ROW_SIZE*MATRIX_ROW_SIZE);//MATRIX_ROW_SIZE);
	system("/home/mathiasb/sources/benchmarking/launch.sh");
	usleep(100000);
	
	WR32(REG_RESET_DMA, 1);
	usleep(100000);
	WR32(REG_RESET_DMA, 0);
	usleep(100000);
	WR32_sleep(REG_PERF_COUNTER,0);
	WR32_sleep(REG_NUM_PACKETS_PER_DESCRIPTOR,64);
	WR32_sleep(REG_PACKET_LENGTH,0x80000 | 64);
	WR32_sleep(REG_TIMER_THRESHOLD, 0x1);
	WR32_sleep(REG_UPDATE_THRESHOLD, 0x1);
	WR64_sleep(REG_UPDATE_COUNTER,descInfo.bus_addr);
	WR64_sleep(REG_UPDATE_ADDRESS,pciBusDesc+DESCRIPTOR_OFFSET);	
        WR32_sleep(REG_CONTROL,CONTROL_ENABLE_READ|CONTROL_SOURCE_RX_FIFO);
	WR32(REG_DMA,1);
	WR32(REG_INTERCONNECT, 0x232);
	WR32(REG_COUNTER,0x1);
	*(int*)gpuDescVa=0;
	WR64(REG_DESCRIPTOR_ADDRESS,pageInfo.bus_addr+i*MATRIX_ROW_SIZE*sizeof(float));
	while(*(int*)gpuDescVa == 0);
	usleep(100000);
    }

    assert_cuda( cudaMemcpy(I.elements,dev_matx.elements,I.rows*I.columns*sizeof(float),cudaMemcpyDeviceToHost) );
    check_identity_matrix(I);

   
    assert_cuda( cudaEventRecord(start) );
    kern_mult_matrix_shared<<< blocks_per_grid,threads_per_block >>>(dev_I,dev_matx,dev_result);
    err = cudaGetLastError();
    printf("%s: %s\n",cudaGetErrorName(err),cudaGetErrorString(err));
    assert_cuda( cudaDeviceSynchronize() );
    assert_cuda( cudaEventRecord(stop) );
    assert_cuda( cudaEventSynchronize(stop) );
    assert_cuda( cudaEventElapsedTime(&ms_shared,start,stop) );
    assert_cuda( cudaMemcpy(result.elements,dev_result.elements,result.rows*result.columns*sizeof(int),cudaMemcpyDeviceToHost) );
    check_identity_matrix(result);
    //check_matrix(result, MATRIX_VALUE);
    printf("Time for GPU (shared): %f\n", ms_shared);

    
    printf("\nClosing the connections\n");
    WR32(REG_COUNTER, 0);
    WR32(REG_DMA, 0);
    WR32(REG_RESET_DMA, 1);
    usleep (100000);
    WR32(REG_RESET_DMA, 0);
    usleep (100000);    
    pcilib_free_kernel_memory(pciCtx,pciHandleDesc,PCILIB_KMEM_FLAG_FORCE);
    pcilib_free_kernel_memory(pciCtx,pciHandlePage,PCILIB_KMEM_FLAG_FORCE);
    assert_cu( cuMemHostUnregister((void*) pciVa) );
    pcilib_close(pciCtx);
    assert_gdr( gdr_unmap(g,GPUMemHandlePage,gpuPageVa,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float)) );
    assert_gdr( gdr_unpin_buffer(g,GPUMemHandlePage) );
    assert_gdr( gdr_unmap(g,GPUMemHandleDesc,gpuDescVa,GPU_PAGE*sizeof(float)) );
    assert_gdr( gdr_unpin_buffer(g,GPUMemHandleDesc) );
    assert_gdr( gdr_close(g) );
    assert_cu( cuMemFree(gpuPagePtr) );
    assert_cu( cuMemFree(gpuDescPtr) );
    assert_cu( cuCtxDestroy(cuCtx) );
    assert_cuda( cudaFree(dev_I.elements) );
    assert_cuda( cudaFree(dev_result.elements) );
    free(I.elements);
    /* free(matx.elements); */
    free(result.elements);
    assert_cuda( cudaEventDestroy(start) );
    assert_cuda( cudaEventDestroy(stop) );

    printf("All Cleared\n");
    exit( EXIT_SUCCESS );
}
