/* This files contains useful fonctions used by all the programs */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "kernels.h"

int stop_process = 0;

void __assert_cuda(cudaError_t err_id, const char* file, int line)
{
    if( err_id != cudaSuccess )
    {
	printf("%s at %s:%i\n",cudaGetErrorString(err_id),file,line);
	exit(EXIT_FAILURE);
    }
}

void __assert_cu(CUresult res_id, const char* file, int line)
{
    if( res_id != CUDA_SUCCESS )
    {
	printf("Error in driver api returned with code: %d at %s:%i\n",res_id,file,line);
	exit(EXIT_FAILURE);
    }
}

void __assert_gdr(int gdr_id, const char* file, int line)
{
    if( gdr_id != 0 )
    {
	printf("Error in gdr api returned with code: %d at %s:%i\n",gdr_id,file,line);
	exit(EXIT_FAILURE);
    }
}

void init_to_send(const void* dataPtr, size_t size, size_t nmemb)
{
    FILE* filePtr = fopen("/home/mathiasb/sources/benchmarking/to_send","wb");
    if( filePtr == NULL )
    {
	printf("Could not open to_send file. Exiting...\n");
	exit( EXIT_FAILURE );
    }
    int errCheck;
    errCheck = fwrite(dataPtr,size,nmemb,filePtr);
    if( errCheck == 0 )
    {
	printf("Could not write the items. Exiting...\n");
	exit( EXIT_FAILURE );
    }
    fclose(filePtr);
}

bool check_array(float* array, float value, size_t size)
{
    for(int i=0;i<size;i++)
    {
	if(array[i] != value)
	{
	    printf("Value error at %d: got %f, expected %f\n",i,array[i],value);
	    return 0;
	}
    }
    printf("Array checked: no error\n");
    return 1;
}

void deviceInformation(int device)
{
    int* value=(int*)malloc(5*sizeof(int));
    const char * query[]={"Threads per block","Block dim X","Grid dim X","Clock rate","Multiprocessor","Threads per Multiprocessor"};
    assert_cuda( cudaDeviceGetAttribute(value+0,cudaDevAttrMaxThreadsPerBlock,0) );
    assert_cuda( cudaDeviceGetAttribute(value+1,cudaDevAttrMaxBlockDimX,0) );
    assert_cuda( cudaDeviceGetAttribute(value+2,cudaDevAttrMaxGridDimX,0) );
    assert_cuda( cudaDeviceGetAttribute(value+3,cudaDevAttrClockRate,0) );
    assert_cuda( cudaDeviceGetAttribute(value+4,cudaDevAttrMultiProcessorCount,0) );
    assert_cuda( cudaDeviceGetAttribute(value+5,cudaDevAttrMaxThreadsPerMultiProcessor,0) );
    for(int i=0;i<5;i++)
	printf("%s: %d\n",query[i],value[i]);
    free(value);
}

void cpu_fill_array(float* array, float value, size_t size)
{
    for(int i=0;i<size;i++)
	array[i] = value;
}

void cpu_fill_array_random(float* array, size_t size)
{
    srand(time(NULL));
    float value;
    for(int i=0;i<size;i++)
    {
	value = ( (float)rand() / (float)RAND_MAX );
	array[i] = value;
    }
}

float cpu_average(float* array, size_t size)
{
    float sum = 0.;
    for(int i=0;i<size;i++)
	sum+=array[i];
    sum /= size;
    return sum;
}

float cpu_dispersion(float* array, float average, size_t size)
{
    float sum = 0.;
    for(int i=0;i<size;i++)
	sum+=(array[i]-average)*(array[i]-average);
    sum /= size;
    return sum;
}

matrix identity_matrix(size_t row_size)
{
    matrix I ;
    I.rows = row_size;
    I.columns = row_size;
    I.stride = I.columns;
    I.elements = (float*)malloc(I.rows*I.columns*sizeof(float));
    if( I.elements==NULL ){
	printf("Cannot create Identity matrix\n");
	exit( EXIT_FAILURE );
    }
    for(int i=0;i<I.rows;i++){
	for(int j=0;j<I.columns;j++){
	    if( i==j)
		I.elements[i+j*I.rows] = 1;
	    else
		I.elements[i+j*I.rows] = 0;
	}
    }
    return I;
}
	
void check_identity_matrix(matrix M)
{
    for(int i=0;i<M.rows;i++){
	for(int j=0;j<M.rows;j++){
	    /* printf("%1.0f-",M.elements[i+j*M.rows]); */
	    if( (i==j && M.elements[i+j*M.rows]!=1) || (i!=j && M.elements[i+j*M.rows]!=0) ){
	    	printf("Error in identity matrix at [%d][%d]\n",i,j);
		exit( EXIT_FAILURE );
	    	}
	}
	/* printf("\n"); */
    }
    printf("Matrix checked: no error\n");
}

void check_matrix(matrix M, int value)
{
    for(int i=0;i<M.rows;i++){
	for(int j=0;j<M.columns;j++){
	    /* printf("%d",M.elements[i*M.columns+j]); */
	    if(M.elements[i*M.columns+j]!=value){
		printf("Error in matrix at [%d][%d]\n",i,j);
		exit( EXIT_FAILURE );
	    }
	}
	/* printf("\n"); */
    }
    printf("Matrix checked: no error\n");
}

void mult_matrix(matrix A, matrix B, matrix R)
{
    int sum;
    for(int i=0;i<A.rows;i++){
        for(int j=0;j<B.columns;j++){
	    sum = 0;
            for(int k=0;k<A.columns;k++){
		sum += A.elements[i*A.columns+k] * B.elements[k*B.columns+j];
            }
	    R.elements[i*R.columns+j] = sum;
        }
    }
}


void fill_matrix_random(matrix M)
{
    srand(time(NULL));
    int value;
    for(int i=0;i<M.rows;i++){
        for(int j=0;j<M.columns;j++){
	    value = (int)rand();
	    M.elements[i*M.columns+j] = value;
	}
    }
}

void* cpu_load_compute(void* arg)
{
    matrix I,matx,result;
    I.rows = MATRIX_ROW_SIZE;
    I.columns = I.rows;
    I.stride = I.columns;
    matx.rows = MATRIX_ROW_SIZE;
    matx.columns = matx.rows;
    matx.stride = matx.columns;
    result.rows = MATRIX_ROW_SIZE;
    result.columns = result.rows;
    result.stride = result.columns;
    I.elements = (float*)malloc(I.rows*I.columns*sizeof(float));
    matx.elements = (float*)malloc(matx.rows*matx.columns*sizeof(float));
    result.elements = (float*)malloc(result.rows*result.columns*sizeof(float));

    fill_matrix_random(I);
    fill_matrix_random(matx);
    
    while(!stop_process){
	mult_matrix(I, matx, result);
    }
    free(I.elements);
    free(result.elements);
    free(matx.elements);
    return NULL;
}

void* gpu_load_compute(void* arg)
{
    dim3 blocks_per_grid(MATRIX_ROW_SIZE/BLOCK_SIZE,MATRIX_ROW_SIZE/BLOCK_SIZE);
    dim3 threads_per_block(BLOCK_SIZE,BLOCK_SIZE);
    matrix dev_I,dev_matx,dev_result;

    dev_I.rows = MATRIX_ROW_SIZE;
    dev_I.columns = dev_I.rows;
    dev_I.stride = dev_I.columns;
    dev_matx.rows = MATRIX_ROW_SIZE;
    dev_matx.columns = dev_matx.rows;
    dev_matx.rows = dev_matx.columns;
    dev_result.rows = MATRIX_ROW_SIZE;
    dev_result.columns = dev_result.rows;
    dev_result.stride = dev_result.columns;
    assert_cuda( cudaMalloc((void**)&dev_I.elements,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float)) );
    assert_cuda( cudaMalloc((void**)&dev_matx.elements,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float)) );
    assert_cuda( cudaMalloc((void**)&dev_result.elements,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(float)) );

    kern_identity_matrix<<< blocks_per_grid,threads_per_block >>>(dev_I);
    kern_identity_matrix<<< blocks_per_grid,threads_per_block >>>(dev_matx);

    while(!stop_process){
	kern_mult_matrix_shared<<< blocks_per_grid,threads_per_block >>>(dev_I,dev_matx,dev_result);
    }
    assert_cuda( cudaFree(dev_I.elements) );
    assert_cuda( cudaFree(dev_result.elements) );
    assert_cuda( cudaFree(dev_matx.elements) );
    return NULL;
}

void* cpu_load_memory(void* arg)
{
    char* foo = (char*) malloc( MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(char) );
    char* bar = (char*) malloc( MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(char) );
    while(!stop_process){
	memcpy(foo, bar, MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(char));
    }
    free(foo);
    free(bar);
    return NULL;
}

void* gpu_load_memory(void* arg)
{
    char *dev_foo,*dev_bar;
    assert_cuda( cudaMalloc((void**)&dev_foo,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(char)) );
    assert_cuda( cudaMalloc((void**)&dev_bar,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(char)) );
    while(!stop_process){
	assert_cuda( cudaMemcpy(dev_foo,dev_bar,MATRIX_ROW_SIZE*MATRIX_ROW_SIZE*sizeof(char),cudaMemcpyDeviceToDevice) );
    }
    assert_cuda( cudaFree(dev_foo) );
    assert_cuda( cudaFree(dev_bar) );
    return NULL;
}

