/* This file contains the kernels i.e. the functions to be executed on the GPU */

#include <stdio.h>
#include <stdlib.h>
#include "kernels.h"




__global__
void fill_array(int* array, int value, size_t size)
{
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint step = blockDim.x*gridDim.x;
    for(int i=0;x+i<size;i+=step)
	array[x+i] = value;
}


__global__
void fill_float_array(float* array, float value, size_t size)
{
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint step = blockDim.x*gridDim.x;
    for(int i=0;x+i<size;i+=step)
	array[x+i] = value;
}


extern __shared__ float _temp[];


__global__ 
void get_reduction(float* array, float* output, size_t size, size_t stride)
{
    float* temp = (float*)_temp;
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    if(x<size)
	temp[threadIdx.x]=array[x];
    __syncthreads();
    for(uint s=1;s<blockDim.x;s*=2)
    {
	x = threadIdx.x*2*s;
	if(x+s < blockDim.x)
	    temp[x]+=temp[x+s];
	__syncthreads();
    }
    if(threadIdx.x==0)
    	output[blockIdx.x+stride*gridDim.x]=temp[0];
}


__global__ 
void get_average_reduction(float* array, float* average, size_t size, size_t stride, size_t divider)
{
    float* temp = (float*)_temp;
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    if(x<size)
	temp[threadIdx.x]=array[x];
    __syncthreads();
    for(uint s=1;s<blockDim.x;s*=2)
    {
	x = threadIdx.x*2*s;
	if(x+s < blockDim.x)
	    temp[x]+=temp[x+s];
	__syncthreads();
    }
    if(threadIdx.x==0)
    	*average=temp[0]/divider;
}


__global__ 
void get_dispersion_reduction(float* array, float* average, float* output, size_t size, size_t stride, size_t divider)
{
    float* temp = (float*)_temp;
    uint x = (blockIdx.x*blockDim.x)+threadIdx.x;
    if(x<size)
	temp[threadIdx.x]=(array[x]-*average) * (array[x]-*average);
    __syncthreads();
    for(uint s=1;s<blockDim.x;s*=2)
    {
	x = threadIdx.x*2*s;
	if(x+s < blockDim.x)
	    temp[x]+=temp[x+s];
	__syncthreads();
    }
    if(threadIdx.x==0)
    	output[blockIdx.x+stride*gridDim.x]=temp[0]/divider;
}


void get_average(float* array, float* average, size_t size,dim3 blocks_per_grid, dim3 threads_per_block)
{
    float *output,*extend;
    size_t stride = size/(blocks_per_grid.x*threads_per_block.x);
    size_t output_size = stride*blocks_per_grid.x;
    size_t extend_size = stride;
    assert_cuda( cudaMalloc((void**)&output,output_size*sizeof(float)) );
    assert_cuda( cudaMalloc((void**)&extend,extend_size*sizeof(float)) );
    stride=0;
    while(stride*blocks_per_grid.x*threads_per_block.x < size)
    {
	get_reduction
	    <<< blocks_per_grid,threads_per_block,threads_per_block.x*sizeof(float)  >>>
	    (array+stride*blocks_per_grid.x*threads_per_block.x,
	     output,
	     blocks_per_grid.x*threads_per_block.x,
	     stride);
	stride++;
    }
    assert_cuda( cudaDeviceSynchronize() );
    get_reduction
	<<< extend_size,blocks_per_grid,blocks_per_grid.x*sizeof(float)  >>>
	(output,
	 extend,
	 output_size,
	 0);
    assert_cuda( cudaDeviceSynchronize() );
    get_average_reduction
	<<< 1,extend_size,extend_size*sizeof(float)  >>>
	(extend,
	 average,
	 extend_size,
	 0,
	 size);
    assert_cuda( cudaFree(output) );
    assert_cuda( cudaFree(extend) );
}


void get_dispersion(float* array, float* average, float* dispersion, size_t size, dim3 blocks_per_grid, dim3 threads_per_block)
{
    get_average(array,average,size,blocks_per_grid,threads_per_block);
    float *output,*extend;
    size_t stride = size/(blocks_per_grid.x*threads_per_block.x);
    size_t output_size = stride*blocks_per_grid.x;
    size_t extend_size = stride;
    assert_cuda( cudaMalloc((void**)&output,output_size*sizeof(float)) );
    assert_cuda( cudaMalloc((void**)&extend,extend_size*sizeof(float)) );
    stride=0;
    while(stride*blocks_per_grid.x*threads_per_block.x < size)
    {
	get_dispersion_reduction
	    <<< blocks_per_grid,threads_per_block,threads_per_block.x*sizeof(float)  >>>
	    (array+stride*blocks_per_grid.x*threads_per_block.x,
	     average,
	     output,
	     blocks_per_grid.x*threads_per_block.x,
	     stride,
	     size);
	stride++;
    }
    assert_cuda( cudaDeviceSynchronize() );
    get_reduction
	<<< extend_size,blocks_per_grid,blocks_per_grid.x*sizeof(float)  >>>
	(output,
	 extend,
	 output_size,
	 0);
    assert_cuda( cudaDeviceSynchronize() );
    get_reduction
      	<<< 1,extend_size,blocks_per_grid.x*sizeof(float)  >>>
    	(extend,
    	 dispersion,
    	 extend_size,
    	 0);
    assert_cuda( cudaFree(output) );
    assert_cuda( cudaFree(extend) );
}


__global__
void fill_matrix(int* M,int value)
{
    uint x = blockIdx.x*blockDim.x +threadIdx.x;
    uint y = blockIdx.y*blockDim.x +threadIdx.y;
    uint step = blockDim.x*gridDim.x;
    for(int i =0;i<blockDim.x;i+=blockDim.y)
	M[x+(y+i)*step] = value;
}


__global__
void add_matrix(int* A,int* B,int* C)
{
    uint i =(blockIdx.x*blockDim.x)+threadIdx.x;
    uint j =(blockIdx.y*blockDim.y)+threadIdx.y;
    C[i*blockDim.x*gridDim.y+j] = A[i*blockDim.x*gridDim.y+j]+B[i*blockDim.x*gridDim.y+j];
}


__global__
void fill_matrix2(int* M,int value,size_t flat_size)
{
    uint i =(blockIdx.x*blockDim.x)+threadIdx.x;
    if( i<flat_size )
	M[i] = value;
}


__global__
void add_matrix2(int* A,int* B,int* C,size_t flat_size)
{
    uint i =(blockIdx.x*blockDim.x)+threadIdx.x;
    if( i<flat_size )
	C[i] = A[i]+B[i];
}


__global__
void add_matrix_mod(int* B,int* C)
{
    uint x =(blockIdx.x*blockDim.x)+threadIdx.x;
    uint y =(blockIdx.y*blockDim.x)+threadIdx.y;
    uint step = blockDim.x*gridDim.x;
    for(int i=0;i<blockDim.x;i+=blockDim.y)
	C[x+(y+i)*step]+=B[x+(y+i)*step];
}


__global__ 
void kern_identity_matrix(matrix I)
{
    uint x = blockIdx.x*blockDim.x +threadIdx.x;
    uint y = blockIdx.y*blockDim.y +threadIdx.y;
    if( x>=I.columns || y>=I.rows )
	return;
    I.elements[x+y*I.columns] = (x==y)? 1:0;
}


__global__
void kern_mult_matrix_naive(matrix A, matrix B, matrix R)
{
    int sum = 0;
    uint col =(blockIdx.x*blockDim.x)+threadIdx.x;
    uint row =(blockIdx.y*blockDim.y)+threadIdx.y;
    if( row > A.rows || col > B.columns ) return;
    for(int k=0;k<R.rows;k++)
	sum += A.elements[row*A.columns+k] * B.elements[k*B.columns+col];
    R.elements[row*R.columns+col] = sum;
}


__global__
void kern_mult_matrix_shared(matrix A, matrix B, matrix R)
{
    int sum = 0;
    uint col = (blockIdx.x*blockDim.x)+threadIdx.x;
    uint row = (blockIdx.y*blockDim.y)+threadIdx.y;
    if( row > A.rows || col > B.columns ) return;

    for(int m=0;m<gridDim.x;m++){
	matrix subA;
	matrix subB;
	subA.stride = A.stride;
	subB.stride = B.stride;
	subA.elements = A.elements + subA.stride*BLOCK_SIZE*blockIdx.y + m*BLOCK_SIZE;
	subB.elements = B.elements + subB.stride*BLOCK_SIZE*m + blockIdx.x*BLOCK_SIZE;
	
	__shared__ int shared_subA[BLOCK_SIZE][BLOCK_SIZE];
	__shared__ int shared_subB[BLOCK_SIZE][BLOCK_SIZE];
	
	shared_subA[threadIdx.y][threadIdx.x] = subA.elements[threadIdx.y*subA.stride+threadIdx.x];
	shared_subB[threadIdx.y][threadIdx.x] = subB.elements[threadIdx.y*subB.stride+threadIdx.x];
	__syncthreads();
	for(int k=0;k<BLOCK_SIZE;k++){
	    sum += shared_subA[threadIdx.y][k] * shared_subB[k][threadIdx.x];
	    __syncthreads();
	}
    }
    R.elements[((blockIdx.y*blockDim.y)+threadIdx.y)*R.stride+(blockIdx.x*blockDim.x)+threadIdx.x] = sum;
}
